/*
 * Xmodem Send / Receive Module
 * adapted from generic Xmodem send/receive code
 * posted to SIG*ATARI on CompuServe
 *
 *	ST users @ Case Western Reserve University
 */

#include <osbind.h>
#include <setjmp.h>
#include "xmdm.h"

 	/* Xmodem specific defines */
#define DOTS		50	/* SECTOR COUNTING DOTS PER LINE */
#define	SECSIZ		0x80	/* Size of a xmodem sector/block */
#define	ERRORMAX	20	/* MAX ERRORS BEFORE ABORT very conservative */
#define	RETRYMAX	15	/* MAXIMUM RETRYS BEFORE ABORT */
#define	SOH		1	/* START OF SECTOR CHAR */
#define	EOT		4	/* end of transmission char */
#define	ACK		6	/* acknowledge sector transmission */
#define	NAK		21	/* error in transmission detected */
#define CAN		('X'&037) /* Cancel transmission	  */


 	/* External Variables */
extern char bufr[ ];	/* file buffer */
extern jmp_buf abort_env,	  /* Long jump when transfers are aborted */
        time_env;		  /* Long jump on timeout		  */

#ifdef DEBUG
extern int debug;
#endif

 	/* Global Variables */
static int mtimeout,	         /* Time out period		   */
	   mdm_sp,     		 /* #packets sent successfully     */
	   mdm_rp,     		 /* #packets received successfully */
	   mdm_packs,     	 /* #packets received/sent	   */
	   mdm_dup,     	 /* #duplicate Packets		   */
	   mdm_bad_packets,	 /* #bad packets      		   */
	   mdm_timeout,		 /* #Timeouts      		   */
	   mdm_naked_packets;	 /* #naked packets    		   */

static long mdm_bytes_xferred;   /* #bytes sent/received sucessfully */




/*
 * mdmini() - initialize variables for a Xmodem Transfer
 *
 */
void mdmini()
{
	mtimeout = 10;
	mdm_bytes_xferred = 0L;
	mdm_sp = 0;
	mdm_rp = 0;
	mdm_bad_packets = 0;
	mdm_timeout = 0;
	mdm_naked_packets = 0;
	mdm_packs = 0;
	mdm_dup = 0;
	
}


/*
 * readfile() - Xmodem Receive
 * We use extremely conservative timeout/retry parameters
 * to deal with lazy dogs like CompuServe. The program
 * could be enhanced to be adaptive by increasing the
 * timeout as and when timeouts occur, dynamically.
 */
int readfile(file)
char *file;
{
	int firstchar,	/* First character of the packet		*/
	sectnum,	/* The currently expected sector # 		*/
	sectcurr,	/* The received sector # 			*/
	sectcomp,	/* The 1's complements of sectcurr		*/
 	fd,		/* File descriptor for opened file	  	*/
	errors,	/* The number of errors encountered in this transfer 	*/
	errorflag;	/* Just a flag					*/
	
	unsigned int
	    checksum,	/* Computed Checksum (of the received data)	*/
	    j,		/* Counter					*/
	    bufptr;	/* Buffer index					*/

	
	/* Initialize myself */
	mdmini ();
	
	/* Create/Open file for write - overwrite if it exists */
	if ((fd = Fcreate(file,0)) < 0) /* Will fail is file present */
	{
		/* Overwrite existing file */
		if((fd = Fopen(file,1)) < 0)
		{
			/* CAN the transfer if we cannot open file */
			Cconws("Cannot Open ");
			Cconws(file);
			Cconws("\r\n");
			canit();
			flushinput();
			return(FALSE);
		}
	}
	else
	{
		/* We are ready now */
		s_display(RECV,XMODEM,file);
		u_message("Receive in Progress .....");
	}
	sectnum = errors = bufptr = 0;
	
    r_again:
	flushinput();
	
	/* Send NAK to let sender know that we are ready */
	sendchar(NAK);
	
	if (setjmp(time_env))
	{
		/* On a Timeout */
		if(++errors < ERRORMAX)
		{
			u_timeouts(++mdm_timeout);
			goto r_again;	/* Sorry about that !! */
		}
		u_message("\007Too many time outs. Aborted Transfer.\007");
		e_display();
		
		Fclose(fd);
		alarm(0);
		canit();
		flushinput();
		return(FALSE);
	}
	
	if (setjmp(abort_env))
	{
		/* User initiated Abort - (CTRL-C) */
		u_message("\007User Aborted Transfer.\007");
		e_display();
		
		Fclose(fd);
		alarm(0);
		canit();
		flushinput();
		return(FALSE);
	}
	
	firstchar = 0;
	while (firstchar != EOT && errors != ERRORMAX)
	{
		errorflag = FALSE;
		alarm (mtimeout);			/* set timeout trap */
		do					/* get sync char */
		{
			firstchar = read_modem();
#ifdef DEBUG
			if (debug)
			{
				Cconws("Got char ");
				printi( firstchar);
				Cconws("\r\n");
			}
#endif
		} while (firstchar != SOH && firstchar != EOT && firstchar != CAN);
		
		if( firstchar == CAN)
		{
			/* Sender Cancelled */
			u_message("\007Transfer Cancelled By Sender.\007");
			e_display();
			
			Fclose(fd);
			flushinput();
			alarm(0);
			return(FALSE);
		}
		
		if (firstchar == SOH)
		{
			/* Ok, got Start of a new packet, pick up the sector # and ~# */
			sectcurr = read_modem();
			sectcomp = read_modem();
			
			u_packets(++mdm_packs);
			
#ifdef DEBUG
			if (debug)
			{
				Cconws("Sector ");
				printi( sectcurr);
				Cconout(' ');
				printi( sectcomp);
				Cconws("\r\n");
			}

#endif
			if ((sectcurr + sectcomp) == 255)
			{
				/* Valid sector # */
				if (sectcurr == (sectnum + 1 & 0xff))
				{
					/* And sector in expected sequence */
					checksum = 0;
					for (j = bufptr;j < (bufptr + SECSIZ);j++)
					{
						/* Pick up data */
						bufr[j] = read_modem();
						/* Computing checksum as we go along */
						checksum = (checksum + bufr[j]) & 0xff;
					}
					if (checksum == read_modem())
					{
						/* Computed Checksum matches received checksum */
						mdm_rp++;
						
						errors = 0;
						sectnum++;
						bufptr += SECSIZ;
						mdm_bytes_xferred += SECSIZ;
						u_bytes(mdm_bytes_xferred);
						
						if (bufptr == BUFSIZ)
						{
							/* Buffer Full - flush it out to file */
							bufptr = 0;
							if (Fwrite(fd,(long)BUFSIZ,bufr) < 0)
							{
								u_message("\007Error writing file - Aborting Transfer.\007");
								e_display();
								
								Fclose(fd);
								alarm(0);
								canit();
								flushinput();
								return(FALSE);
							};
						};
						flushinput ();
						/* Every thing ok - send the ACK */
						sendchar(ACK);
					}
					else
					{
						/* Received Checksum doe not match computed checksum */
						u_errors(++mdm_bad_packets);
						
						
						errorflag = TRUE;
#ifdef DEBUG
						if (debug)
						{
							Cconout('C');
						};
#endif
					}
				}
				else
				{
					/* Sector received is not the expected one */
					/* Could be a duplicate sector, or just a curropted sector */
					if (sectcurr == (sectnum & 0xff))
					{
						/* Duplicate sector - */
						/* just ignore it and ACK */
						u_dup(++mdm_dup);
						
						flushinput();
						sendchar(ACK);
					}
					else
					{
						/* Garbage */
#ifdef DEBUG
						if (debug)
						{
							Cconws("synch error\r\n");
						}
#endif
						
						u_errors(++mdm_bad_packets);
						
						errorflag = TRUE;
					}
				}
			}
			else
			{
#ifdef DEBUG
				if (debug)
				{
					Cconws("sector number error\r\n");
				}
#endif
				u_errors(++mdm_bad_packets);
				
				errorflag = TRUE;
			}
		}

		if (errorflag == TRUE)
		{
			/* NAK it if errorflag is set */
			errors++;
			flushinput();
			sendchar(NAK);
		}
		
	};	/* end while  Not EOT and errors < ERRORMAX */
	
	
	if ((firstchar == EOT) && (errors < ERRORMAX))
	{
		/* Last packet */
		sendchar(ACK);
		if(Fwrite(fd,(long)bufptr,bufr) != bufptr)
		{
			u_message("\007Error writing file.\007");
			e_display();
			
			Fclose(fd);
			alarm(0);
			return(FALSE);
		}
		Fclose(fd);
		alarm(0);
		ui_message(mdm_bytes_xferred," Bytes Received.\007");
		e_display();
		
		return(TRUE);
	}
	alarm(0);
	e_display();
	
	return(FALSE);
}

/*
 * sendfile() - Xmodem Send
 * See comments for readfile() concerning timeouts/retries.
 *
 */
int sendfile(file)
char *file;
{
	int sectnum,	/* Current sector number */
	sectors,	/* #of sectors sent      */
	attempts,	/* #of attempts to send  */
	fd;		/* File descriptor	 */
	
	unsigned int
	    checksum,	/* Computed checksum	 */
	j,		/* all purpose counter	 */
	bufptr;		/* Ptr to file buffer	 */
	
	char
	    c;		/* General Character var */
	
	/* init. myself */
	mdmini ();

	/* Open file for read */
	if ((fd = Fopen(file, 0)) < 0)
	{
		/* CAN tranfer if we can't open */
		Cconws("cannot open ");
		Cconws(file);
		Cconws("\r\n");
		canit();
		flushinput();
		return(FALSE);
	}
	else
	{
		s_display(SEND,XMODEM,file);
		u_message("Awaiting Receiver .....");
	}

	/* All set to go */
	attempts = 0;
	sectnum = 1;
	j = 0;
	
	
	if (setjmp(abort_env))
	{
		/* On user initiated abort (CTRL-C) */
		u_message("\007User Aborted Transfer.\007");
		e_display();
		
		Fclose(fd);
		alarm(0);
		canit();
		flushinput();
		return(FALSE);
	}
	
	attempts = 0;
	
	do {
	    s_again:
		
		if (setjmp(time_env))
		{
			/* On Time Out */
			if(attempts < (2*ERRORMAX))
			{
				u_timeouts(++mdm_timeout);
				flushinput();
				goto s_again;	/* Sorry about that !! */
			}
			u_message("\007Too many time outs. Aborted Transfer.\007");
			e_display();

			Fclose(fd);
			alarm(0);
			canit();
			flushinput();
			return(FALSE);
		}
		
		/* Wait for First char from receiver */
		alarm(mtimeout);
		c = read_modem();
		
#ifdef DEBUG
		if (debug)
		{
			Cconws("Ate char ");
			printi( c);
			Cconws("\r\n");
		}
#endif
		/* Eat everything but a NAK or CAN */
	} while ((c != NAK) && (c != CAN) && (j++ < (ERRORMAX*2)));
	
	if( c == CAN)
	{
		/* receiver cancelled */
		u_message("\007Transfer Cancelled By Receiver.\007");
		e_display();
		
		Fclose(fd);
		flushinput();
		alarm(0);
		return(FALSE);
	}
	
	if (j >= (ERRORMAX*2))
	{
		/* Too many errors */
		u_message("\007Receiver not sending NAK's - Aborting.\007");
		e_display();
		
		canit();
		flushinput();
		alarm(0);
		return(FALSE);
	}
	
#ifdef DEBUG
	if (debug)
	    Cconws ("Got ACK\r\n");
#endif
	u_message("Send in Progress .....");
	flushinput (); /* Get rid of any junk on the line */
	alarm(mtimeout);
	
	while ((sectors = Fread(fd,(long)BUFSIZ,bufr))
	       && (attempts != RETRYMAX))
	{
		if (sectors < 0)
		{
			/* Unable to read the file */
			u_message("\007Error reading file - Aborting Transfer.\007");
			e_display();
			
			Fclose(fd);
			alarm(0);
			canit();
			flushinput();
			return(FALSE);
		}

		bufptr = 0;
		do
		{
			attempts = 0;
			do
			{
			    s_ag2:
				alarm(mtimeout);
				if (setjmp(time_env))
				{
					/* on time out */
					if(attempts < (2*ERRORMAX))
					{
						u_timeouts(++mdm_timeout);
						flushinput();
						goto s_ag2;	/* Sorry about that !! */
					}
					u_message("\007Too many time outs. Aborted Transfer.\007");
					e_display();

					Fclose(fd);
					alarm(0);
					canit();
					flushinput();
					return(FALSE);
				}
				
				u_packets(++mdm_packs);
				
				/* Send a packet */
				sendchar(SOH);
				sendchar(sectnum);
				sendchar(~sectnum);
				checksum = 0;
				write_modem(&bufr[bufptr],SECSIZ);
				for (j = bufptr; j < (bufptr+SECSIZ); j++)
				    checksum += bufr[j];
				sendchar(checksum);
				flushinput();
				mdm_sp++;
				
				attempts++;
				/* Did it get sent over ? */
				c = read_modem();
				
				if (c != ACK)
				    /* Oh no */
				    u_errors(++mdm_naked_packets);
#ifdef DEBUG
				if (debug)
				{
					Cconws ("Response char is ");
					printi( c);
					Cconws("\r\n");
				}
#endif
				if( c == CAN )
				{ 
					/* Receiver cancelled - so i'll abort */
					u_message("\007Transfer Cancelled By Receiver.\007");
					e_display();
					
					Fclose(fd);
					flushinput();
					alarm(0);
					return(FALSE);
				}
				
			} while ((c != ACK) && (attempts != RETRYMAX));
			/* Packet got sent and I received an Ack, or
			   too many attempts */
			bufptr += SECSIZ;
			mdm_bytes_xferred += SECSIZ;
			sectnum++;
		} while ((bufptr < sectors) && (attempts != RETRYMAX));
	};
	
	Fclose(fd);
	if (attempts == RETRYMAX)
	{
		/* Too many attempts */
		u_message("\007No acknowledgment of sector, Aborting.\007");
		e_display();
		alarm(0);
		canit();
		flushinput();
		return(FALSE);
	}
	else
	{
		/* All sectors that needed to be sent are done,
		   Now send the EOT to complete transfer,
		   The EOT may or may not get ACK'ed, warn
		   the user if it does'nt */
		attempts = 0;
		do
		{
		    s_ag3:
			alarm(mtimeout);
			if (setjmp(time_env))
			{
				if(attempts < (2*ERRORMAX))
				{
					u_timeouts(++mdm_timeout);
					flushinput();
					goto s_ag3;	/* Sorry about that !! */
				}
			}
			
			sendchar(EOT);
			attempts++;
		} while ((read_modem() != ACK) && (attempts != RETRYMAX));

		alarm(0);
		if (attempts == RETRYMAX)
		{
		    ui_message(mdm_bytes_xferred,
		    " Bytes Sent. No acknowledgment of EOF - Hopefully the file made it.\007");
		}
		else
		{
		    ui_message(mdm_bytes_xferred," Bytes Sent.\007");
		}
			
	};

	e_display();
	return(TRUE);
}


/*
 * canit() - Cancel the transfer
 */

void canit()
{
	register int i;
	
	/* Send 5 CAN's */
	for(i = 0; i < 5; i++)
	    sendchar(CAN);

	/* Get rid of 4 of them, some receivers expect at least 2
	   CAN to confirm that the first CAN was not a corrupted
	   ACK, but eat only one (Ymodem for instance). */
	   
	for(i = 0; i < 4; i++)
	    sendchar('\b');
}
