/* dlink - data link transfer routines for bmodem */
/*
	by David Betz, BYTE Magazine/BIX
*/

#include <stdio.h>

/* useful definitions */
#define TRUE	1
#define FALSE	0

/* protocol characters */
#define SOH	0x01	/* start of a 128 byte block */
#define STX	0x02	/* start of a 1024 byte block */
#define EOT	0x04	/* end of a complete file */
#define ACK	0x06	/* positive acknowledgement */
#define NAK	0x15	/* negative acknowledgement */
#define CAN	0x18	/* cancel transfer */
#define CRC	'C'	/* CRC request (instead of NAK to start a transfer) */

/* routine status codes */
#define DT_EOF	-1
#define DT_ERR	-2
#define DT_OK	1
#define DT_TIME	-1

/* data message length */
#define SMSGLEN	128	/* short message length */
#define LMSGLEN	1024	/* long message length */

/* number of times to retry */
#define RETRY	10

/* timeout loop counters */
#define STIME	1	/* timeout between characters of a packet */
#define LTIME	10	/* timeout waiting for an ACK/NAK */
#define XTIME	60	/* timeout waiting for initial NAK */

/* the data buffer (big enough for long messages) */
static unsigned char packet[LMSGLEN+5];
static unsigned char *buffer = &packet[3];	/* data portion of packet */

/* global variables */
static long fsize;	/* total bytes transfered */
static int blknum;	/* current block number */
static int bufptr;	/* current buffer pointer */
static int chksum;	/* current checksum value (or CRC) */
static int msgstart;	/* current message start character (SOH or STX) */
static int msglength;	/* current message length */
static int crcmode;	/* TRUE for CRC mode, FALSE for checksum */
static int nak;		/* NAK character on receive */

/* dl_snd - send data across the link */
int dl_snd(getch,usecrc,uselarge)
  int (*getch)(),usecrc,uselarge;
{
    int ch,retries;

    /* setup message length */
    if (uselarge) {
	msgstart = STX;
	msglength = LMSGLEN;
    }
    else {
	msgstart = SOH;
	msglength = SMSGLEN;
    }

    /* initialize */
    fsize = 0L;		/* initialize the total file size */
    blknum = 1;		/* start a block number 1 */
    bufptr = 0;		/* start with an empty buffer */

    /* wait for the initial NAK (or CRC) */
    for (retries = 0; retries < RETRY; retries++)
	if ((ch = md_get(XTIME)) == NAK) {	/* start of checksum transfer */
	    crcmode = FALSE;
	    break;
	}
	else if (usecrc && ch == CRC) {		/* start of CRC transfer */
	    crcmode = TRUE;
	    break;
	}
	else
	    dlabort();

    /* check for failure */
    if (retries >= RETRY)
	return (FALSE);

    /* send each byte */
    while ((ch = (*getch)()) != EOF)
	if (msgput(ch) == DT_ERR)
	    return (FALSE);

    /* flush partial buffer */
    if (bufptr > 0)
	while (bufptr > 0)
	    if (msgput(0) == DT_ERR)
		return (FALSE);

    /* send EOT and wait for ACK */
    for (retries = 0; retries < RETRY; retries++) {
	md_put(EOT);
	md_flush();
	if (md_get(LTIME) == ACK)
	    return (TRUE);
    }

    /* return failure */
    return (FALSE);
}

/* dl_rcv - receive data across the link */
int dl_rcv(putch,usecrc)
  int (*putch)(),usecrc;
{
    int ch;

    /* initialize */
    fsize = 0L;
    blknum = 1;
    bufptr = msglength = SMSGLEN;
    
    /* send the initial NAK (or CRC) */
    nak = (usecrc ? CRC : NAK);
    crcmode = usecrc;
    md_put(nak);
    md_flush();

    /* receive each byte */
    while ((ch = msgget()) != DT_EOF && ch != DT_ERR)
	(*putch)(ch);

    /* return with status */
    return (ch == DT_EOF);
}

/* msgput - put a message data character */
static int msgput(ch)
  int ch;
{
    int sts;

    buffer[bufptr++] = ch;
    if (bufptr == msglength) {
	fsize += (long)msglength;
	printf("\rSending %d, %ld bytes",blknum,fsize); fflush(stdout);
	if ((sts = mdsnd()) != DT_OK)
	    return (sts);
	bufptr = 0;
	++blknum;
    }
    return (ch);
}

/* msgget - get a message data character */
static int msgget()
{
    int sts;

    if (bufptr == msglength) {
	if ((sts = mdrcv()) != DT_OK)
	    return (sts);
	fsize += (long)msglength;
	printf("\rReceived %d, %ld bytes",blknum,fsize); fflush(stdout);
	bufptr = 0;
	++blknum;
    }
    return ((int)buffer[bufptr++]);
}

/* mdsnd - send a buffer of data */
static int mdsnd()
{
    int plength,retries,ch;

    /* setup the packet header */
    packet[0] = msgstart;
    packet[1] = blknum;
    packet[2] = ~blknum;

    /* compute the block check code */
    if (crcmode) {				/* compute the CRC */
	chksum = crc_compute(buffer,msglength);
	buffer[msglength] = chksum >> 8;
	buffer[msglength+1] = chksum;
	plength = msglength + 5;
    }
    else {					/* compute the checksum */
	chksum = chk_compute(buffer,msglength);
	buffer[msglength] = chksum;
	plength = msglength + 4;
    }

    /* send data and wait for an ACK */
    for (retries = 0; retries < RETRY; retries++) {

	/* send the data */
	md_write(packet,plength);
	md_flush();

tryagain:
	/* return on an ACK */
	if ((ch = md_get(LTIME)) == ACK)
	    return (DT_OK);

	/* abort transfer on two successive CAN's */
	else if (ch == CAN) {
	    if ((ch = md_get(STIME)) == ACK)
		return (DT_OK);
	    else if (ch == CAN)
		return (DT_ERR);
	    else if (ch != NAK)
		goto tryagain;
	}

	/* ignore all other characters */
	else if (ch != NAK)
	    goto tryagain;
    }

    /* return failure */
    return (DT_ERR);
}

/* mdrcv - receive a buffer of data */
static int mdrcv()
{
    int plength,i,ch,retries;

    /* receive data */
    for (retries = RETRY; retries; --retries) {

	/* check for data packet or eot */
	if ((ch = md_get(LTIME)) == DT_TIME) {
	    if (crcmode && blknum == 1 && retries <= RETRY-2) {
		printf("    switching to checksum mode"); fflush(stdout);
		crcmode = FALSE;
		nak = NAK;
	    }
	    md_put(nak);
	    continue;
	}
	else if (ch == EOT) {		/* end of transfer */
	    md_put(ACK);
	    md_flush();
	    return (DT_EOF);
	}
	else if (ch == SOH)		/* start of a short packet */
	    msglength = SMSGLEN;
	else if (ch == STX)		/* start of a long packet */
	    msglength = LMSGLEN;
	else {
	    dlabort();
	    if (crcmode && blknum == 1 && retries <= RETRY-2) {
		printf("    switching to checksum mode"); fflush(stdout);
		crcmode = FALSE;
		nak = NAK;
	    }
	    md_put(nak);
	    continue;
	}

	/* reset the NAK character */
	nak = NAK;

	/* compute the packet length */
	plength = msglength + (crcmode ? 5 : 4);

	/* receive the data */
	for (i = 1; i < plength; packet[i++] = ch)
	    if ((ch = md_get(STIME)) == DT_TIME)
		break;

	/* check for timeout */
	if (i < plength) {
	    md_put(NAK);
	    md_flush();
	    continue;
	}

	/* check the block number */
	if (packet[1] != (~packet[2] & 0xFF)) {
	    md_put(NAK);
	    md_flush();
	    continue;
	}

	/* check the block check code */
	if (crcmode) {					/* CRC */
	    chksum = (buffer[msglength] << 8) | buffer[msglength+1];
	    if (chksum != crc_compute(buffer,msglength)) {
		md_put(NAK);
		md_flush();
		continue;
	    }
	}
	else {						/* checksum */
	    chksum = buffer[msglength];
	    if (chksum != chk_compute(buffer,msglength)) {
		md_put(NAK);
		md_flush();
		continue;
	    }
	}

	/* check the block number */
	if (packet[1] == (blknum & 0xFF)) {
	    md_put(ACK);
	    md_flush();
	    return (DT_OK);
	}
	else if (packet[1] != ((blknum - 1) & 0xFF)) {
	    md_put(CAN);
	    md_put(CAN);
	    md_flush();
	    break;
	}
	
	/* send a ack */
	md_put(ACK);
	md_flush();
    }

    /* return failure */
    return (DT_ERR);
}

/* dlabort - wait for the line to clear */
static dlabort()
{
    while (md_get(STIME) != DT_TIME)
	;
}

/* chk_compute - compute the checksum */
static int chk_compute(buf,len)
  unsigned char *buf;
  int len;
{
    int chk,i;

    for (chk = i = 0; i < len; ++i)
	chk += buf[i];
    return (chk & 0xFF);
}

/* crc_compute - compute the CRC value */
static int crc_compute(buf,len)
  unsigned char *buf;
  int len;
{
    int crc,shifter,highbit,i;

    buf[len++] = '\0'; buf[len++] = '\0';
    for (crc = i = 0; i < len; ++i)
	for (shifter = 0x80; shifter; shifter >>= 1) {
	    highbit = crc & 0x8000;
	    crc <<= 1;
	    crc |= (buf[i] & shifter ? 1 : 0);
	    if (highbit) crc ^= 0x1021;
	}
    return (crc & 0xFFFF);
}
